import json
import os
from typing import List
from unittest import TestCase

import yaml

import octoploy.octoploy
from octoploy.config.Config import RunMode, RootConfig
from octoploy.state.StateMover import StateMover
from octoploy.state.StateTracking import StateTracking, ObjectState
from tests import TestUtils
from tests.TestUtils import DummyK8sApi


class StateTrackingTest(TestCase):

    def setUp(self) -> None:
        self._base_path = os.path.dirname(__file__)
        self._tmp_file = 'out.yml'
        self._mode = RunMode()
        self._dummy_api = DummyK8sApi()
        octoploy.octoploy.load_project = self._load_project

    def tearDown(self) -> None:
        self._dummy_api = None

    def _load_project(self, config_dir: str) -> RootConfig:
        def get_dummy_api():
            return self._dummy_api

        prj_config = RootConfig.load(os.path.join(self._base_path, config_dir))
        prj_config.get_state()._k8s_api = self._dummy_api
        prj_config.create_api = get_dummy_api
        return prj_config

    def _create_state_data(self):
        data = {}
        obj = ObjectState()
        obj.update_from_key('a/b/c.d/12')
        data[obj.get_key()] = obj
        self.assertEqual('a', obj.context)
        self.assertEqual('b', obj.namespace)
        self.assertEqual('c.d/12', obj.fqn)

        obj = ObjectState()
        obj.update_from_key('a/b/c')
        data[obj.get_key()] = obj
        self.assertEqual('a', obj.context)
        self.assertEqual('b', obj.namespace)
        self.assertEqual('c', obj.fqn)

        obj = ObjectState()
        obj.update_from_key('a/b/Deployment/name')
        data[obj.get_key()] = obj
        self.assertEqual('a', obj.context)
        self.assertEqual('b', obj.namespace)
        return data

    def test_move(self):
        root = self._load_project('app_deploy_test')
        mover = StateMover(root)
        state = StateTracking(root.create_api())
        data = state._state
        root._state = state

        def void(ignore):
            pass

        root.initialize_state = void
        data.update(self._create_state_data())

        self.assertEqual('a', data['a/b/c.d/12'].context)
        self.assertEqual('b', data['a/b/c.d/12'].namespace)
        self.assertEqual('c.d/12', data['a/b/c.d/12'].fqn)

        self.assertEqual('a', data['a/b/c'].context)
        self.assertEqual('b', data['a/b/c'].namespace)
        self.assertEqual('c', data['a/b/c'].fqn)

        self.assertEqual('a', data['a/b/Deployment/name'].context)
        self.assertEqual('b', data['a/b/Deployment/name'].namespace)

        mover.move('a/b/c.d/12', '1/2/3/4', None)
        self.assertEqual('1', data['1/2/3/4'].context)
        self.assertEqual('2', data['1/2/3/4'].namespace)
        self.assertEqual('3/4', data['1/2/3/4'].fqn)
        data = state._state = self._create_state_data()

        mover.move('a/b', '1/2', None)
        self.assertEqual('1', data['1/2/Deployment/name'].context)
        self.assertEqual('2', data['1/2/Deployment/name'].namespace)
        self.assertEqual('Deployment/name', data['1/2/Deployment/name'].fqn)

        self.assertEqual('1', data['1/2/c'].context)
        self.assertEqual('2', data['1/2/c'].namespace)
        self.assertEqual('c', data['1/2/c'].fqn)
        data = state._state = self._create_state_data()

    def test_create_new(self):
        self._dummy_api.respond(['get', 'Deployment/ABC', '-o', 'json'], '', error=Exception('NotFound'))
        self._dummy_api.respond(['get', 'ConfigMap/octoploy-state', '-o', 'json'], '{"kind": "", "apiVersion": ""}')

        octoploy.octoploy._run_app_deploy('app_deploy_test', 'app', self._mode)

        self.assertEqual(4, len(self._dummy_api.commands))
        self.assertEqual(['get', 'ConfigMap/octoploy-state', '-o', 'json'], self._dummy_api.commands[0].args)
        state_update = self._dummy_api.commands[-1]
        self.assertEqual(['apply', '-f', '-'], state_update.args)
        self.assertStateEqual([{"context": "ABC",
                                "hash": "e2e4634c5cd31a1b58da917e8b181b28",
                                "fqn": "Deployment/ABC",
                                "namespace": "oc-project",
                                }], state_update.stdin)

    def test_duplicate_kinds(self):
        self._dummy_api.respond(['get', 'Deployment/ABC', '-o', 'json'], '', error=Exception('NotFound'))
        self._dummy_api.respond(['get', 'ConfigMap/octoploy-state', '-o', 'json'], '{"kind": "", "apiVersion": ""}')

        octoploy.octoploy._run_app_deploy('app_deploy_test_duplicate_kinds', 'app', self._mode)

        self.assertEqual(4, len(self._dummy_api.commands))
        self.assertEqual(['get', 'ConfigMap/octoploy-state', '-o', 'json'], self._dummy_api.commands[0].args)
        state_update = self._dummy_api.commands[-1]
        self.assertEqual(['apply', '-f', '-'], state_update.args)
        self.assertStateEqual([{"context": "app",
                                "hash": "aa859898df4ff9412857e720beeabfba",
                                "fqn": "ProviderConfig.kubernetes.crossplane.io/default",
                                "namespace": "oc-project",
                                },
                               {"context": "app",
                                "hash": "8652aee0d35b000096f2a263c6e3eb77",
                                "fqn": "ProviderConfig.grafana.crossplane.io/default",
                                "namespace": "oc-project",
                                }], state_update.stdin)

    def test_deploy_all_then_app(self):
        """
        Deploys an entire folder, then a single app and makes
        sure the other objects don't get removed again from k8s
        """
        self._dummy_api.not_found_by_default()

        os.environ['OCTOPLOY_KEY'] = TestUtils.OCTOPLOY_KEY
        octoploy.octoploy._run_apps_deploy('app_deploy_test', self._mode)

        self.assertEqual(14, len(self._dummy_api.commands))
        self.assertEqual(['get', 'ConfigMap/octoploy-state', '-o', 'json'], self._dummy_api.commands[0].args)

        state_update = self._dummy_api.commands[-1]
        self.assertEqual(['apply', '-f', '-'], state_update.args)
        self.assertStateEqual([
            {"context": "ABC",
             "hash": "e2e4634c5cd31a1b58da917e8b181b28",
             "fqn": "Deployment/ABC",
             "namespace": "oc-project"},
            {"context": "entity-compare-api",
             "hash": "644921ab79abf165d8fb8304913ff1c7",
             "fqn": "Deployment/8080",
             "namespace": "oc-project"},
            {"context": "favorite-api",
             "hash": "3005876d55a8c9af38a58a4227a377fc",
             "fqn": "Deployment/8081",
             "namespace": "oc-project"},
            {"context": "cm-types",
             "hash": "1f4d778af0ea594402e656fd6139c584",
             "fqn": "ConfigMap/config",
             "namespace": "oc-project"},
            {"context": "ABC2",
             "hash": "d6e8e8f4b59b77117ec4ad267be8dcae",
             "fqn": "Secret/secret",
             "namespace": "oc-project"},
            {"context": "var-append",
             "hash": "24d921e38f585e26cdc247d8fddc260e",
             "fqn": "ConfigMap/config",
             "namespace": "oc-project"},
        ], state_update.stdin)

        # Now deploy a single app
        current_state = yaml.safe_load(state_update.stdin)
        self._dummy_api.respond(['get', 'ConfigMap/octoploy-state', '-o', 'json'], json.dumps(current_state))
        self._dummy_api.respond(['get', 'Deployment/ABC', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.commands = []
        octoploy.octoploy._run_app_deploy('app_deploy_test', 'app', self._mode)

        self.assertEqual(3, len(self._dummy_api.commands))
        state_update = self._dummy_api.commands[-1]
        new_state = yaml.safe_load(state_update.stdin)
        self.assertEqual(current_state, new_state)

    def test_deploy_all_twice(self):
        """
        Deploys an entire folder twice and validates that the objects don't get removed again from k8s
        """
        self._dummy_api.not_found_by_default()

        os.environ['OCTOPLOY_KEY'] = TestUtils.OCTOPLOY_KEY
        octoploy.octoploy._run_apps_deploy('app_deploy_test', self._mode)

        self.assertEqual(14, len(self._dummy_api.commands))
        self.assertEqual(['get', 'ConfigMap/octoploy-state', '-o', 'json'], self._dummy_api.commands[0].args)

        state_update = self._dummy_api.commands[-1]
        self.assertEqual(['apply', '-f', '-'], state_update.args)
        self.assertStateEqual([{"context": "ABC",
                                "hash": "e2e4634c5cd31a1b58da917e8b181b28",
                                "fqn": "Deployment/ABC",
                                "namespace": "oc-project"},
                               {"context": "entity-compare-api",
                                "hash": "644921ab79abf165d8fb8304913ff1c7",
                                "fqn": "Deployment/8080",
                                "namespace": "oc-project"},
                               {"context": "favorite-api",
                                "hash": "3005876d55a8c9af38a58a4227a377fc",
                                "fqn": "Deployment/8081",
                                "namespace": "oc-project"},
                               {"context": "cm-types",
                                "hash": "1f4d778af0ea594402e656fd6139c584",
                                "fqn": "ConfigMap/config",
                                "namespace": "oc-project"},
                               {"context": "ABC2",
                                "hash": "d6e8e8f4b59b77117ec4ad267be8dcae",
                                "fqn": "Secret/secret",
                                "namespace": "oc-project"},
                               {"context": "var-append",
                                "hash": "24d921e38f585e26cdc247d8fddc260e",
                                "fqn": "ConfigMap/config",
                                "namespace": "oc-project"},
                               ], state_update.stdin)
        # Now deploy a single app
        current_state = yaml.safe_load(state_update.stdin)
        self._dummy_api.respond(['get', 'ConfigMap/octoploy-state', '-o', 'json'], json.dumps(current_state))
        self._dummy_api.respond(['get', 'Deployment/ABC', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.respond(['get', 'Deployment/8080', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.respond(['get', 'Deployment/8081', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.respond(['get', 'ConfigMap/config', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.respond(['get', 'Secret/secret', '-o', 'json'], '{"kind": "", "apiVersion": ""}')
        self._dummy_api.commands = []
        octoploy.octoploy._run_apps_deploy('app_deploy_test', self._mode)

        self.assertEqual(8, len(self._dummy_api.commands))
        for x in range(7):
            self.assertEqual('get', self._dummy_api.commands[x].args[0])

        state_update = self._dummy_api.commands[-1]
        new_state = yaml.safe_load(state_update.stdin)
        self.assertEqual(current_state, new_state)

    def test_removed_in_repo(self):
        self._dummy_api.respond(['get', 'DeploymentConfig/ABC', '-o', 'json'], '', error=Exception('NotFound'))
        self._dummy_api.respond(['get', 'ConfigMap/octoploy-state', '-o', 'json'], '''{
  "kind": "test",
  "apiVersion": "v1",
  "data": {
    "state": "[{\\"context\\": \\"ABC\\", \\"fqn\\": \\"DeploymentConfig/ABC\\", \\"namespace\\": \\"oc-project\\"}]"
  },
  "fqn": "ConfigMap",
  "metadata": {
    "name": "octoploy-state"
  }
}''')

        octoploy.octoploy._run_app_deploy('app_deploy_test_empty', 'app', self._mode)

        self.assertEqual(3, len(self._dummy_api.commands))
        self.assertEqual(['get', 'ConfigMap/octoploy-state', '-o', 'json'], self._dummy_api.commands[0].args)
        self.assertEqual(['delete', 'DeploymentConfig/ABC'], self._dummy_api.commands[1].args)

        state_update = self._dummy_api.commands[2]
        self.assertEqual(['apply', '-f', '-'], state_update.args)
        self.assertStateEqual([], state_update.stdin)

    def assertStateEqual(self, expected: List[any], data: str):
        k8s_object = yaml.safe_load(data)
        state = yaml.safe_load(k8s_object['data']['state'])
        self.assertEqual(len(expected), len(state))
        for entry in expected:
            found_match = False
            for existing in state:
                if entry == existing:
                    found_match = True
                    break
            self.assertTrue(found_match, f'{entry} not found in {state}')



from unittest import TestCase
from unittest.mock import patch

from octoploy.k8s.BaseObj import BaseObj
from octoploy.k8s.K8sObjectDiff import K8sObjectDiff
from octoploy.utils.Log import ColorFormatter
from tests.TestUtils import DummyK8sApi


class K8sObjectDiffTest(TestCase):

    @patch('builtins.print')
    def test_secret_masking(self, mock_print):
        api = DummyK8sApi()
        a = BaseObj({
            'kind': 'Secret',
            'apiVersion': 'v1',
            'spec': {
                'data': '1',
            }
        })
        b = BaseObj({
            'kind': 'Secret',
            'apiVersion': 'v1',
            'spec': {
                'data': '2',
                'stringData': 'newField',
            }
        })

        K8sObjectDiff(api).print(a, b)

        stdout_lines = ''
        for args in mock_print.call_args_list:
            stdout_lines += ColorFormatter.decolorize(str(args.args[0]))+'\n'

        self.assertNotIn('1', stdout_lines)
        self.assertNotIn('2', stdout_lines)
        self.assertIn('data', stdout_lines)
        self.assertIn('stringData', stdout_lines)
        self.assertNotIn('newField', stdout_lines)


def test_diff(self):
    api = DummyK8sApi()
    a = BaseObj({
        'kind': 'a',
        'apiVersion': 'v1',
        'metadata': {
            'annotations': {
                'kubectl.kubernetes.io/last-applied-configuration': 'ignore',
            },
            'uid': '',
            'resourceVersion': '123',
        },
        'spec': {
            'field': 'value',
            'removed': 'value',
            'listItemAdd': ['a'],
            'listRemoved': ['a', 'b'],
            'listItemRemoved': ['a', 'b'],
            'listItemChanged': ['a', 'b'],
            'removedDict': {'a': 'hello'}
        }
    })
    b = BaseObj({
        'kind': 'a',
        'apiVersion': 'v1',
        'metadata': {
            'annotations': {
                'newAnnotation': 'hello',
            },
        },
        'spec': {
            'field': 'otherValue',
            'listItemAdd': ['a', 'b'],
            'listAdded': ['a', 'b'],
            'listItemRemoved': ['a'],
            'listItemChanged': ['a', 'other'],
            'addedDict': {'a': 'hello'}
        }
    })

    K8sObjectDiff(api).print(a, b)


def test_multiline_str(self):
    api = DummyK8sApi()
    a = BaseObj({
        'kind': 'a',
        'apiVersion': 'v1',
        'spec': {
            'field': '''
                This is a multiline string
                It has multiple lines
                Yay''',
        }
    })
    b = BaseObj({
        'kind': 'a',
        'apiVersion': 'v1',
        'spec': {
            'field': '''
                This is a multiline string
                It has multiple lines, but some have been changed
                or have been added
                Yay''',
        }
    })

    K8sObjectDiff(api).print(a, b)



from __future__ import annotations

import argparse

from octoploy.backup.BackupGenerator import BackupGenerator
from octoploy.config.Config import RootConfig, RunMode
from octoploy.converter.HelmToOcto import HelmToOcto
from octoploy.deploy.AppDeploy import AppDeployment
from octoploy.processing.DecryptionProcessor import DecryptionProcessor
from octoploy.state.StateMover import StateMover
from octoploy.utils.Encryption import YmlEncrypter
from octoploy.utils.Log import Log

log_instance = Log('octoploy')


def load_project(config_dir: str) -> RootConfig:
    if config_dir != '':
        return RootConfig.load(config_dir)

    # No path specified, try a few common ones
    paths = ['.', 'configs', 'octoploy']
    for path in paths:
        try:
            return RootConfig.load(path)
        except FileNotFoundError:
            continue
    raise FileNotFoundError(f'Did not find config in any of {paths}')


def reload_config(args):
    root_config = load_project(args.config_dir)
    app_config = root_config.load_app_config(args.name[0])
    if not app_config.enabled():
        log_instance.log.error('App is disabled')
        return
    if app_config.is_template():
        log_instance.log.error('App is a template')
        return

    oc = root_config.create_api()
    log_instance.log.info('Reloading ' + app_config.get_name())
    reload_actions = app_config.get_reload_actions()
    for action in reload_actions:
        action.run(oc)
    log_instance.log.info('Done')


def _run_app_deploy(config_dir: str, app_name: str, mode: RunMode):
    root_config = load_project(config_dir)
    root_config.initialize_state(mode)
    app_config = root_config.load_app_config(app_name)
    try:
        deployment = AppDeployment(root_config, app_config, mode)
        deployment.deploy()
    finally:
        root_config.persist_state(mode)
    log_instance.log.info('Done')


def _run_apps_deploy(config_dir: str, mode: RunMode):
    root_config = load_project(config_dir)
    root_config.initialize_state(mode)
    configs = root_config.load_app_configs()
    log_instance.log.debug(f'Found {len(configs)} apps to deploy')
    try:
        for app_config in configs:
            AppDeployment(root_config, app_config, mode).deploy()
    finally:
        root_config.persist_state(mode)
    log_instance.log.info('Done')


def plan_app(args):
    mode = RunMode()
    mode.plan = True
    mode.set_override_env(args.env)
    _run_app_deploy(args.config_dir, args.name[0], mode)


def deploy_app(args):
    mode = RunMode()
    mode.out_file = args.out_file
    mode.dry_run = args.dry_run
    mode.set_override_env(args.env)
    _run_app_deploy(args.config_dir, args.name[0], mode)


def delete_app(args):
    mode = RunMode()
    mode.delete = True
    mode.plan = args.plan
    mode.set_override_env(args.env)
    _run_app_deploy(args.config_dir, args.name[0], mode)


def plan_all(args):
    mode = RunMode()
    mode.plan = True
    mode.set_override_env(args.env)
    _run_apps_deploy(args.config_dir, mode)


def deploy_all(args):
    mode = RunMode()
    mode.out_file = args.out_file
    mode.dry_run = args.dry_run
    mode.set_override_env(args.env)
    _run_apps_deploy(args.config_dir, mode)


def create_backup(args):
    root_config = load_project(args.config_dir)
    BackupGenerator(root_config).create_backup(args.name[0])


def convert_helm(args):
    dest = args.config_dir
    if dest == '':
        dest = '.'
    converter = HelmToOcto(dest)
    if args.filter is not None:
        for f in args.filter:
            converter.include(f)
    converter.convert(args.file)


def encrypt_secrets(args):
    files = args.file
    for file in files:
        YmlEncrypter(file).encrypt()


def list_state(args):
    root_config = load_project(args.config_dir)
    root_config.initialize_state(RunMode())

    state = root_config.get_state()
    state.print()


def move_state(args):
    source = args.items[0]
    dest = args.items[1]
    target_cm = args.to

    root_config = load_project(args.config_dir)
    mover = StateMover(root_config)
    mover.move(source, dest, dest_configmap=target_cm)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', dest='version', action='store_true',
                        help='Prints the current version')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Enables debug logging')
    parser.add_argument('--skip-secrets', dest='skip_secrets', action='store_true',
                        help="Skips all secret objects and therefore doesn't require a key to be set")
    parser.add_argument('--deploy-plain-secrets', dest='deploy_plain_text', action='store_true',
                        help="Deploys plain text secret objects")
    parser.add_argument('-c', '--config-dir', dest='config_dir',
                        help='Path to the folder containing all configurations',
                        default='')
    parser.add_argument('-e', '--env', dest='env', action='append',
                        help='key=value pairs of parameters to set/override during for the templating engine',
                        default=[])

    subparsers = parser.add_subparsers(help='Commands')
    state_parser = subparsers.add_parser('state', help='State interactions')

    state_subparsers = state_parser.add_subparsers(help='State commands')
    state_list_parser = state_subparsers.add_parser('list')
    state_list_parser.set_defaults(func=list_state)

    state_mv_parser = state_subparsers.add_parser('mv')
    state_mv_parser.set_defaults(func=move_state)

    state_mv_parser.add_argument('items', help='Source and destination.\n'
                                               'The format for an object is: octoployName/Namespace/k8sFqn\n'
                                               'For example: "my-old-app/Namespace2/k8sFqn" '
                                               '"my-new-app/Namespace2/k8sFqn"', nargs=2)
    state_mv_parser.add_argument('--to', dest='to', help='Configmap where the state should be moved to (optional).'
                                                         'Format: configmapSuffix|namespace/configmapSuffix')
    state_mv_parser.set_defaults(func=move_state)

    helm_parser = subparsers.add_parser('helm-convert', help='Converts a rendered helm chart yml file to octoploy')
    helm_parser.add_argument('file', help='Yml file to be converted')
    helm_parser.add_argument('--filter', dest='filter', help='Component names to be converted - optional', nargs='*')
    helm_parser.set_defaults(func=convert_helm)

    encrypt_parser = subparsers.add_parser('encrypt', help='Encrypts k8s secrets objects')
    encrypt_parser.add_argument('file', help='Yml file to be encrypted', nargs=1)
    encrypt_parser.set_defaults(func=encrypt_secrets)

    backup_parser = subparsers.add_parser('backup', help='Creates a backup of all resources in the cluster')
    backup_parser.add_argument('name', help='Name of the backup folder', nargs=1)
    backup_parser.set_defaults(func=create_backup)

    reload_parser = subparsers.add_parser('reload', help='Reloads the configuration of a running application')
    reload_parser.add_argument('name', help='Name of the app which should be reloaded (folder name)', nargs=1)
    reload_parser.set_defaults(func=reload_config)

    plan_parser = subparsers.add_parser('plan', help='Verifies what changes have to be applied for a single app')
    plan_parser.add_argument('name', help='Name of the app which should be checked (folder name)', nargs=1)
    plan_parser.set_defaults(func=plan_app)

    plan_all_parser = subparsers.add_parser('plan-all',
                                            help='Verifies what changes have to be applied for all apps')
    plan_all_parser.set_defaults(func=plan_all)

    deploy_parser = subparsers.add_parser('deploy', help='Deploys the configuration of an application')
    deploy_parser.add_argument('--out-file', dest='out_file',
                               help='Writes all objects into a yml file instead of deploying them. '
                                    'This does not communicate with openshift in any way')
    deploy_parser.add_argument('--dry-run', dest='dry_run', help='Does not interact with k8s/openshift '
                                                                 '(pure template processing). '
                                                                 'Can be used in conjunction with out-file to preview '
                                                                 'templates',
                               action='store_true')
    deploy_parser.add_argument('name', help='Name of the app which should be deployed (folder name)', nargs=1)
    deploy_parser.set_defaults(func=deploy_app)

    deploy_all_parser = subparsers.add_parser('deploy-all',
                                              help='Deploys all objects of all enabled application')
    deploy_all_parser.add_argument('--out-file', dest='out_file',
                                   help='Writes all objects into a yml file instead of deploying them. '
                                        'This does not communicate with openshift in any way')
    deploy_all_parser.add_argument('--dry-run', dest='dry_run', help='Does not interact with openshift',
                                   action='store_true')
    deploy_all_parser.set_defaults(func=deploy_all)

    delete_parser = subparsers.add_parser('delete', help='Deletes the configuration of an application')
    delete_parser.add_argument('--plan', dest='plan', help='Same as plan but in delete mode',
                               action='store_true')
    delete_parser.add_argument('name', help='Name of the app which should be deleted', nargs=1)
    delete_parser.set_defaults(func=delete_app)

    args = parser.parse_args()
    if args.version:
        from octoploy import __version__
        print(f'Octoploy {__version__}')
        return

    if 'func' not in args.__dict__:
        parser.print_help()
        exit(1)
        return

    if args.debug:
        Log.set_debug()

    DecryptionProcessor.skip_secrets = args.skip_secrets
    DecryptionProcessor.deploy_plain_text = args.deploy_plain_text
    args.func(args)


if __name__ == '__main__':
    main()



from abc import abstractmethod
from typing import Dict, Optional

from octoploy.k8s.BaseObj import BaseObj


class TreeProcessor:
    @abstractmethod
    def process(self, k8s_object: BaseObj):
        """
        Processes the tree

        :param k8s_object: K8s object
        :raises: SkipObject: The object should be skipped
        """
        pass

    def process_object(self, data: Dict[str, any], parent: Dict[str, any], key: str) -> Optional[Dict[str, any]]:
        """
       Processes a node of the tree
       :param data: Object
       :param parent: Parent holding the object
       :param key: Key of the object
       :return: New value or None if the parent should not be updated
       """
        return data

    def process_str(self, value: str, parent: Dict[str, any], key: str) -> Optional[any]:
        """
        Processes a string leaf of the tree
        :param value: Value
        :param parent: Parent holding the value
        :param key: Key of the value
        :return: New value or None if the value should not be updated
        """
        return value


class TreeWalker:
    """
    Walks through dictionary trees
    """

    def __init__(self, processor: TreeProcessor):
        self.processor = processor

    def walk(self, data: Dict[str, any]):
        self._walk(data)

    def _walk(self, data: Dict[str, any]):
        for key in list(data.keys()):
            value = data[key]
            new_val = self._process_item(value, data, key)
            if new_val is None:
                continue
            data[key] = new_val

    def _process_item(self, value, data, key) -> any:
        if isinstance(value, list):
            for idx, list_item in enumerate(value):
                value[idx] = self._process_item(list_item, data, key)
            return value

        if isinstance(value, str):
            return self.processor.process_str(value, data, key)

        if isinstance(value, dict):
            new_val = self.processor.process_object(value, data, key)
            if new_val is None:
                return None
            self._walk(new_val)
            return new_val
        return value



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from octoploy.config.Config import RootConfig

from octoploy.k8s.BaseObj import BaseObj
from octoploy.processing.TreeWalker import TreeProcessor


class NamespaceProcessor(TreeProcessor):
    _root: RootConfig

    def __init__(self, root: RootConfig):
        self._root = root

    def process(self, k8s_object: BaseObj):
        self._update_namespace(k8s_object)

    def _update_namespace(self, k8s_object: BaseObj):
        """
        Updates the namespace meta-data of this object.
        If no namespace is defined this will use the globaly defined namespace of the octoploy project
        :param k8s_object: Object
        """
        # An object might be in a different namespace than the current context
        namespace = k8s_object.namespace
        if namespace is None:
            namespace = self._root.get_namespace_name()
        k8s_object.set_namespace(namespace)  # Make sure the object points to the correct namespace



from typing import Dict

from octoploy.k8s.BaseObj import BaseObj


class SecretObj(BaseObj):
    def __init__(self, data: Dict[str, any]):
        super().__init__(data)
        self.require_kind('secret')
        self.base64_data = data.get('data', {})
        self.string_data = data.get('stringData', {})




from typing import Optional, Tuple

from octoploy.config.Config import RootConfig, RunMode
from octoploy.state.StateTracking import StateTracking
from octoploy.utils.Log import Log


class StateMover(Log):
    def __init__(self, source_config: RootConfig):
        super().__init__(__name__)
        self._sourceConfig = source_config

    def move(self, source: str, dest: str, dest_configmap: Optional[str]):
        """
        Moves the state from the given source to the given destination
        :param source: Source
        :param dest: Destination
        :param dest_configmap: Different configmap which should receive the state
        """
        if source.count('/') != dest.count('/'):
            raise ValueError('Source and destination point to different path depths')

        run_mode = RunMode()
        self._sourceConfig.initialize_state(run_mode)
        source_state = self._sourceConfig.get_state()

        items_to_be_moved = source_state.get_items(source)
        for item in items_to_be_moved:
            key = item.get_key()
            target = key.replace(source, dest)
            self.log.info(f'Moving {key} to {target}')
            item.update_from_key(target)
            source_state.remove_key(key)

        target_state = source_state
        target_namespace = self._sourceConfig.get_namespace_name()
        if dest_configmap is not None:
            target_state, target_namespace = self._get_state_from_cm(dest_configmap)

        for item in items_to_be_moved:
            target_state.add(item)

        if len(items_to_be_moved) == 0:
            self.log.warning('No items moved')
            return

        # First persist the target to prevent data loss on error
        if target_state != source_state:
            target_state.store(target_namespace)
        source_state.store(self._sourceConfig.get_namespace_name())

    def _get_state_from_cm(self, configmap: str) -> Tuple[StateTracking, str]:
        namespace = self._sourceConfig.get_namespace_name()
        segments = configmap.split('/', 1)
        if len(segments) == 2:
            namespace = segments[0]
            cm_suffix = segments[1]
        else:
            cm_suffix = segments[0]

        api = self._sourceConfig.create_api()
        state = StateTracking(api, name_suffix=cm_suffix)
        state.restore(namespace)
        return state, namespace



from typing import Dict, Optional

from octoploy.k8s.BaseObj import BaseObj

from octoploy.k8s.SecretObj import SecretObj
from octoploy.processing.TreeWalker import TreeProcessor, TreeWalker
from octoploy.utils import Utils
from octoploy.utils.Encryption import Encryption
from octoploy.utils.Errors import SkipObject
from octoploy.utils.Log import Log


class DecryptionProcessor(TreeProcessor, Log):
    """
    Processes all objects, and replaces any encrypted placeholders.
    If skip_secrets is True, all secret object types wil lbe skipped
    """

    skip_secrets: bool = False
    """
    Indicates if all secrets should be skipped
    """

    deploy_plain_text: bool = False
    """
    Indicates if plain text secrets should be skipped
    """

    _secret_obj: Optional[SecretObj]

    def __init__(self):
        super().__init__(__name__)
        self.encryption = Encryption()

    def process(self, k8s_object: BaseObj):
        try:
            self._secret_obj = SecretObj(k8s_object.data)
        except ValueError:
            self._secret_obj = None

        walker = TreeWalker(self)
        walker.walk(k8s_object.data)

    def process_str(self, value: str, parent: Dict[str, any], key: str) -> str:
        if not value.startswith(Encryption.CRYPT_PREFIX):
            if self._secret_obj is None:
                # Not a secret object, simply pass through
                return value

            # The secret may contain non encrypted values
            # We'll ignore them for now since we don't want to change the behavior
            # of octoploy for existing projects.
            # Also, it may increase chances of accidentally storing them in vcs
            if key in self._secret_obj.base64_data or \
                    key in self._secret_obj.string_data:
                if not DecryptionProcessor.deploy_plain_text:
                    raise SkipObject(
                        f'Secret {self._secret_obj.get_fqn()} contains plain text - '
                        f'use "octoploy encrypt" to encrypt your secrets')

            return value

        if DecryptionProcessor.skip_secrets and self._secret_obj is not None:
            # We should skip the secret
            raise SkipObject('Secrets should be skipped')

        payload = Utils.remove_prefix(value, Encryption.CRYPT_PREFIX)
        return self.encryption.decrypt(payload)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from octoploy.k8s.BaseObj import BaseObj
from octoploy.utils.Log import Log

if TYPE_CHECKING:
    from octoploy.config.Config import AppConfig

from octoploy.api.Kubectl import K8sApi


class DeploymentActionConfig(Log):
    """
    Configuration for a single user configurable deployment step
    """

    def __init__(self, app_config: AppConfig, data):
        super().__init__()
        self._data = data
        self._app_config = app_config

    def run(self, k8s: K8sApi, all_objects: List[BaseObj]):
        namespace = self._app_config.get_root().get_namespace_name()
        if self._data == 'deploy':
            deployment_name = self._app_config.get_name()
            deployment_obj = self.get_deployment_object(deployment_name, all_objects)
            if deployment_obj is None:
                self.log.warning(f'Deployment object {deployment_name} not found')
                return

            try:
                k8s.rollout(deployment_name, namespace=deployment_obj.namespace)
            except Exception as e:
                if '(NotFound)' in str(e):
                    self.log.warning(f'Could not restart {deployment_name} {deployment_obj.namespace}:\n{e}')
                    return
            return

        exec_config = self._data.get('exec', None)
        if exec_config is not None:
            cmd = exec_config['command']
            args = exec_config['args']

            dc_name = self._app_config.get_name()
            self.log.info('Reloading via exec in pods of ' + dc_name)
            pods = k8s.get_pods(dc_name=dc_name, namespace=namespace)
            for pod in pods:
                k8s.exec(pod.name, cmd, args, namespace=namespace)
            return

    @staticmethod
    def get_deployment_object(deployment_name: str, all_objects: List[BaseObj]) -> Optional[BaseObj]:
        for obj in all_objects:
            if (obj.is_kind('Deployment') or obj.is_kind('DeploymentConfig')) \
                    and obj.name == deployment_name:
                return obj
        return None



from abc import abstractmethod
from typing import Optional, List

from octoploy.api.Kubectl import K8sApi
from octoploy.k8s.BaseObj import BaseObj


class DeploymentMode:
    """
    Defines how an object should be deployed
    """
    _api: K8sApi
    _flags: List[str]

    def __init__(self):
        self._flags = []

    def use_api(self, api: K8sApi):
        self._api = api

    @abstractmethod
    def deploy(self, k8s_object: BaseObj, existing_object: Optional[BaseObj], namespace: str):
        pass

    def set_flags(self, flags: List[str]):
        """
        Sets additional flags that should be used during deployment
        :param flags: Flags
        """
        self._flags = flags


class ReplaceDeploymentMode(DeploymentMode):
    """
    Replace the object
    """

    def deploy(self, k8s_object: BaseObj, existing_object: Optional[BaseObj], namespace: str):
        self._api.replace(k8s_object.as_string(), namespace=namespace, extra_flags=self._flags)


class ApplyDeploymentMode(DeploymentMode):
    """
    Apply the object
    """

    def deploy(self, k8s_object: BaseObj, existing_object: Optional[BaseObj], namespace: str):
        self._api.apply(k8s_object.as_string(), namespace=namespace, extra_flags=self._flags)



from __future__ import annotations

from typing import Dict, List, Optional

import yaml

from octoploy.api.Kubectl import K8sApi
from octoploy.k8s.BaseObj import BaseObj
from octoploy.utils.Log import Log
from octoploy.utils.YmlWriter import YmlWriter


class ObjectState:
    fqn: str
    """
    Fully qualified name of the object in the format
    kind.group/name
    """

    visited: bool
    """
    Transient flag to indicate if the object has been visited by octoploy
    """

    def __init__(self):
        self.context: str = ''
        self.namespace: str = ''

        self.fqn = ''

        self.hash = ''
        self.visited = False

    def update_from_key(self, key: str):
        segments = key.split('/', 2)
        if len(segments) < 3:
            raise ValueError(f'Invalid key, must be 3 segments long {key}')

        self.context = segments[0]
        self.namespace = segments[1]
        self.fqn = segments[2]

    def parse(self, data: Dict[str, str]) -> ObjectState:
        self.context = data.get('context')
        self.namespace = data.get('namespace')
        self.fqn = data.get('fqn')
        if self.context is None or self.namespace is None or self.fqn is None:
            raise ValueError(f'Corrupt octoploy state, could not parse {data}')

        self.hash = data.get('hash', '')
        return self

    def to_dict(self) -> Dict[str, str]:
        return {
            'context': self.context,
            'namespace': self.namespace,
            'fqn': self.fqn,
            'hash': self.hash,
        }

    def get_key(self) -> str:
        return f'{self.context}/{self.namespace}/{self.fqn}'


class StateTracking(Log):
    """
    Stores the objects that have been deployed with octoploy.
    This allows octoploy to detect renamed / deleted objects.
    """
    CM_NAME = 'octoploy-state'
    _k8s_api: K8sApi
    _state: Dict[str, ObjectState]

    def __init__(self, api: K8sApi, name_suffix: str = ''):
        super().__init__()
        self._k8s_api = api
        self._cm_name = self.CM_NAME + name_suffix
        self._state = {}

    def restore(self, namespace: str):
        item = self._k8s_api.get(f'ConfigMap/{self._cm_name}', namespace=namespace)
        if item is None:
            return
        cm = item.data
        state_data_str = cm.get('data', {}).get('state', '')
        state_data = yaml.safe_load(state_data_str)
        if state_data is None:
            return

        for state_obj in state_data:
            object_state = ObjectState().parse(state_obj)
            self._state[object_state.get_key()] = object_state

    def store(self, namespace: str):
        self.log.debug(f'Persisting state in ConfigMap {self._cm_name}')

        states = []
        for object_state in self._state.values():
            states.append(object_state.to_dict())

        data = {
            'kind': 'ConfigMap',
            'apiVersion': 'v1',
            'metadata': {
                'name': self._cm_name
            },
            'data': {
                'state': YmlWriter.dump(states)
            }
        }
        yml = YmlWriter.dump(data)
        self._k8s_api.apply(yml, namespace=namespace)

    def add(self, object_state: ObjectState):
        self._state[object_state.get_key()] = object_state

    def remove(self, object_state: ObjectState):
        del self._state[object_state.get_key()]

    def remove_key(self, key: str):
        del self._state[key]

    def get_items(self, prefix: str) -> List[ObjectState]:
        """
        Returns all state items which start with the given prefix
        :param prefix: Prefix
        :return: Items
        """
        items = []
        for value in list(self._state.values()):
            key = value.get_key()
            if not key.startswith(prefix):
                continue
            items.append(value)
        return items

    def get_not_visited(self, context: str) -> List[ObjectState]:
        """
        Returns all objects which have not been visited
        :param context: Context
        :return: Objects
        """
        items = []
        for object_state in self._state.values():
            if not object_state.visited and object_state.context == context:
                items.append(object_state)
        return items

    def get_state(self, context_name: str, k8s_object: BaseObj) -> Optional[ObjectState]:
        state = self._k8s_to_state(context_name, k8s_object)
        return self._state.get(state.get_key())

    def visit(self, context_name: str, k8s_object: BaseObj, hash_val: str, only_update: bool = False):
        """
        Marks the given object as "visited".
        If the object is not yet in the state it will be added
        :param context_name: Name of the state context
        :param k8s_object: Kubernetes object
        :param hash_val: The new hash value of the object.
        :param only_update: True if the state should only be updated and not added if not existing
        """
        state = self._k8s_to_state(context_name, k8s_object)
        existing_state = self._state.get(state.get_key())
        if existing_state is None:
            if only_update:
                return
            state.hash = hash_val
            self._state[state.get_key()] = state
            return
        existing_state.hash = hash_val
        existing_state.visited = True

    def visit_only(self, context_name: str, k8s_object):
        """
        Marks the given object as "visited" if already in the state
        """
        state = self._k8s_to_state(context_name, k8s_object)
        existing_state = self._state.get(state.get_key())
        if existing_state is not None:
            existing_state.visited = True

    def print(self):
        self.log.info(f'State content of ConfigMap {self._cm_name}')
        for key, value in self._state.items():
            self.log.info('|- ' + value.get_key())

    @staticmethod
    def _k8s_to_state(context_name: str, k8s_object: BaseObj) -> ObjectState:
        # At this point we always have a namespace set for the object
        state = ObjectState()
        state.context = context_name
        state.namespace = k8s_object.namespace
        state.fqn = k8s_object.get_fqn()
        state.visited = True
        return state



import os
from unittest import TestCase

from octoploy.utils.Encryption import Encryption


class EncryptionTest(TestCase):

    def test_encrypt_valid(self):
        os.environ['OCTOPLOY_KEY'] = 'key123'
        enc = Encryption()
        out = enc.encrypt('hello world')

        enc = Encryption()
        self.assertEqual('hello world', enc.decrypt(out))

    def test_wrong_key(self):
        os.environ['OCTOPLOY_KEY'] = 'asd123aopiksd'
        enc = Encryption()
        out = enc.encrypt('hello world')

        os.environ['OCTOPLOY_KEY'] = 'wrong key'
        enc = Encryption()
        self.assertRaises(ValueError, enc.decrypt, out)



from __future__ import annotations

from typing import Optional, Dict, List, Set
from typing import TYPE_CHECKING

from octoploy.k8s.BaseObj import BaseObj
from octoploy.processing.TreeWalker import TreeWalker, TreeProcessor
from octoploy.utils.Log import Log

if TYPE_CHECKING:
    from octoploy.config.BaseConfig import BaseConfig
from octoploy.utils.Errors import MissingParam


class YmlTemplateProcessor(Log, TreeProcessor):
    """
    Processes yml files by replacing any string placeholders.
    """

    KEY_FIELD_MERGE: str = '_merge'

    _config: BaseConfig
    _replacements: Dict[str, any]
    _parents: List[YmlTemplateProcessor]
    _child: Optional[YmlTemplateProcessor]

    _missing_vars: List[str]
    """
    List of all variables which have not been replace because
    there was no value defined for them
    """

    def __init__(self, config: BaseConfig):
        super().__init__()
        self._missing_vars = []

        self._config = config
        self._parents = []
        self._child = None
        self._replacements = {}

    def parents(self, template_processor: List[YmlTemplateProcessor]):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the parent, the definition of this processor will override
        the parent definition

        :param template_processor: Child processor
        """
        if len(self._parents) > 0:
            raise ValueError('Parent processors already defined')
        self._parents = template_processor

    def child(self, template_processor: YmlTemplateProcessor):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the child, the definition of the child processor will override
        this definition

        :param template_processor: Child processor
        """
        if self._child is not None:
            raise ValueError('Child processor already defined')
        self._child = template_processor

    def process(self, k8s_object: BaseObj):
        """
        Processes the app data

        :param k8s_object: Data of the app, the data will be modified in place
        :raise MissingParam: Gets raised if at least one parameter is not defined
        """
        self._load_replacements()

        # Now replace any placeholders in the actual data tree
        walker = TreeWalker(self)
        walker.walk(k8s_object.data)
        k8s_object.refresh()

        # Check if any of the missing vars are declared as "params"
        # (aka are required)
        if len(self._missing_vars) > 0:
            missing_params = []
            params = self._get_params()
            for missing in self._missing_vars:
                if missing not in params:
                    continue
                missing_params.append(missing)
            if len(missing_params) > 0:
                raise MissingParam('The following params are not defined: ' + str(missing_params))
            self.log.warning('The following vars are not defined: ' + str(self._missing_vars))

    def _get_params(self) -> Set[str]:
        """
        Returns all defined params
        :return: Param names
        """
        params = set()
        for parent in self._parents:
            params.update(parent._get_params())
        params.update(self._config.get_params())
        if self._child is not None:
            params.update(self._child._get_params())
        return params

    def process_object(self, data: Dict[str, any], parent: Dict[str, any], key: str) -> Optional[Dict[str, any]]:
        if key == self.KEY_FIELD_MERGE:
            # Remove the item from the parent
            # and merge it into the parent itself
            del parent[key]
            TreeWalker(self).walk(data)
            parent.update(data)
            return None
        return data

    def process_str(self, value: str, parent: Dict[str, any], key: str) -> any:
        if key == self.KEY_FIELD_MERGE:
            # Remove the item from the parent
            # and merge it into the parent itself
            del parent[key]
            new_value = self._replace(value)
            if isinstance(new_value, str):
                raise ValueError(f'No replacement found for {new_value}')
            parent.update(new_value)
            return None

        return self._replace(value)

    def _replace(self, item: str) -> any:
        """
        Replaces the given variable with the actual value
        :param item: Variable in the format "${KEY}" "$$" will be escaped to "$"
        :return: Value or item if no replacement was found
        """
        new_item = ''
        var_name_buffer = ''
        mode = 0
        for char in item:
            if mode == 0:  # Search for tag start $
                if char == '$':
                    mode = 1
                    continue
                new_item += char
                continue

            if mode == 1:  # After $
                if char == '$':
                    # Escaped $
                    new_item += '$'
                    mode = 0
                    continue

                if char == '{':
                    mode = 2
                    var_name_buffer = ''
                    continue

                # Not a variable tag, just add the $ and the current char
                new_item += '$' + char
                mode = 0
                continue

            if mode == 2:  # After ${, search for end }
                if char == '}':
                    # End of variable tag found
                    mode = 0
                    new_val = self._replacements.get(var_name_buffer)
                    if new_val is None:
                        self._missing_vars.append(var_name_buffer)
                        new_item += '${' + var_name_buffer + '}'
                        continue

                    # The new_val can be an "object" as well,
                    # but we can only replace it if the given item
                    # only refers to this variable and is not a combination of strings
                    if item == '${' + var_name_buffer + '}':
                        return new_val

                    if not isinstance(new_val, str):
                        raise ValueError(f'Invalid replacement for {var_name_buffer}: Expected string, got {new_val}.\n'
                                         f'Non string replacements are only possible for single variable references.')
                    new_item += new_val
                    continue
                var_name_buffer += char
                continue

        if mode == 1:
            # Ended with a $
            new_item += '$'
        if mode == 2:
            # Ended with a ${, so we need to add the buffer
            new_item += '${' + var_name_buffer

        return new_item

    def _load_replacements(self):
        """
        Loads all available replacements
        """
        self._replacements = self._get_replacements()
        self._resolve_refs(self._replacements)

    def _get_replacements(self) -> Dict[str, any]:
        """
        Returns all replacements handled by this processor, including all parent variables
        :return: Replacements
        """
        replacements = {}
        for parent in self._parents:
            replacements.update(parent._get_replacements())
        replacements.update(self._config.get_replacements())
        if self._child is not None:
            replacements.update(self._child._get_replacements())
        return replacements

    def _resolve_refs(self, data: Dict[str, any]):
        """
        Resolves any references in the given replacements (in place)
        :param data: Replacements
        """
        found_var_to_replace = True
        while found_var_to_replace:
            found_var_to_replace = False
            for key, value in data.items():
                if isinstance(value, dict):
                    # A replacement might be an object
                    # So we need to walk through every item and see if there is something to replace
                    self._resolve_refs(value)
                    continue
                if not isinstance(value, str):
                    continue

                # The replacement value might refer to another variable
                new_val = self._replace(value)
                if new_val != value:
                    data[key] = new_val
                    found_var_to_replace = True
                    continue



from typing import List

from octoploy.api.Kubectl import K8sApi
from octoploy.config.Config import RootConfig, AppConfig, RunMode
from octoploy.k8s.BaseObj import BaseObj
from octoploy.k8s.K8sObjectDiff import K8sObjectDiff
from octoploy.state.StateTracking import StateTracking
from octoploy.utils.Log import Log, ColorFormatter


class K8sObjectDeployer(Log):
    """
    Deploys k9s object
    """

    HASH_ANNOTATION = 'yml-hash'
    """
    Name of the annotation that stores the hash
    """
    _root_config: RootConfig
    _app_config: AppConfig
    _api: K8sApi
    _state: StateTracking

    def __init__(self, root_config: RootConfig, k8sapi: K8sApi, app_config: AppConfig, mode: RunMode = RunMode()):
        super().__init__()
        self._root_config = root_config
        self._app_config = app_config
        self._api = k8sapi
        self._mode = mode
        self._state = root_config.get_state()

        self._to_be_deployed: List[BaseObj] = []

    def add_object(self, k8s_object: BaseObj):
        """
        Adds the given object to the deployment list
        :param k8s_object: Object which should be deployed
        """
        self._to_be_deployed.append(k8s_object)
        self._state.visit_only(self._app_config.get_name(), k8s_object)

    def execute(self):
        """
        Executes the deployment
        """
        self._delete_abandoned_objects()
        self._deploy_objects()

    def _deploy_objects(self):
        """
        Deploys the pending objects
        """
        for k8s_object in self._to_be_deployed:
            self._deploy_object(k8s_object)

    def _deploy_object(self, k8s_object: BaseObj):
        """
        Deploy the given object (if a deployment required, otherwise does nothing)
        :param k8s_object: Object which should be deployed
        """
        hash_val = k8s_object.get_hash()
        item_path = k8s_object.get_fqn()
        namespace = k8s_object.namespace

        current_object = self._api.get(item_path, namespace=namespace)
        if current_object is None:
            if self._mode.delete:
                return
            self._log_create(item_path)

        state_hash = None
        old_state_hash = None
        obj_state = None
        if current_object is not None:
            old_state_hash = current_object.get_annotation(self.HASH_ANNOTATION)
            obj_state = self._state.get_state(self._app_config.get_name(), k8s_object)
            if obj_state is not None and obj_state.hash != '':
                state_hash = obj_state.hash
            else:  # Fallback to old hash location
                state_hash = old_state_hash

        if self._mode.delete:
            if current_object is not None:
                self._log_delete(item_path)
                if self._mode.plan:
                    return
                self._api.delete(item_path, namespace=namespace)
            if obj_state is not None:
                self._state.remove(obj_state)
            return

        if current_object is not None and state_hash is None:
            # Item has not been deployed with octoploy, but it does already exist
            self.log.warning(f'{item_path} has no state, assuming no change required')
            self._state.visit(self._app_config.get_name(), k8s_object, hash_val)
            return

        if state_hash == hash_val:
            self.log.debug(f"{item_path} hasn't changed")
            return

        if current_object is not None:
            self._log_update(item_path)
            if self._mode.plan:
                K8sObjectDiff(self._api).print(current_object, k8s_object)

        if self._mode.plan:
            return

        if old_state_hash is not None:
            # Migrate to new state format by removing the old one
            self._api.annotate(k8s_object.get_fqn(), self.HASH_ANNOTATION, None, namespace=k8s_object.namespace)

        deploy_mode = self._app_config.get_deployment_mode()
        deploy_mode.use_api(self._api)
        deploy_mode.deploy(k8s_object, current_object, namespace=namespace)

        # Update hash
        self._state.visit(self._app_config.get_name(), k8s_object, hash_val)

        if k8s_object.is_kind('ConfigMap'):
            self._reload_config()

    def _delete_abandoned_objects(self):
        """
        Deletes all objects which are not anymore included in the
        current app config
        """
        abandoned = self._state.get_not_visited(self._app_config.get_name())
        for item in abandoned:
            namespace = item.namespace

            self._log_delete(item.fqn)
            if self._mode.plan:
                continue

            self._api.delete(item.fqn, namespace=namespace)
            self._state.remove(item)

    def _reload_config(self):
        """
        Tries to reload the configuration for the app
        """
        reload_actions = self._app_config.get_reload_actions()
        for action in reload_actions:
            action.run(self._api, self._to_be_deployed)

    def _log_create(self, item_path: str):
        self._log_verb(item_path, ColorFormatter.colorize('+', ColorFormatter.green), 'created', 'creating')

    def _log_delete(self, item_path: str):
        self._log_verb(item_path, ColorFormatter.colorize('-', ColorFormatter.red), 'deleted', 'deleting')

    def _log_update(self, item_path: str):
        self._log_verb(item_path, ColorFormatter.colorize('~', ColorFormatter.yellow), 'updated', 'updating')

    def _log_verb(self, item_path: str, symbol: str, past_verb: str, progressive_verb: str):
        if self._mode.plan or self._mode.dry_run:
            self.log.info(f'{symbol} {item_path} will be {past_verb}')
            return
        self.log.info(f'{symbol} {progressive_verb} {item_path}')



import os
from unittest import TestCase

from octoploy.config.BaseConfig import BaseConfig
from octoploy.processing.ValueLoader import ValueLoaderFactory


class ValueLoaderTest(TestCase):

    def test_env(self):
        factory = ValueLoaderFactory()
        loader = factory.create(BaseConfig(None), 'env')
        items = loader.load({})
        self.assertTrue(len(items) > 0)

    def test_file_abspath(self):
        factory = ValueLoaderFactory()
        loader = factory.create(BaseConfig(None), 'file')
        items = loader.load({'file': __file__})
        self.assertIsNotNone(items.get(''))
        self.assertIn('class', items.get(''))

        items = loader.load({'file': __file__, 'conversion': 'base64'})
        self.assertNotIn('class', items.get(''))

    def test_file_relpath(self):
        factory = ValueLoaderFactory()
        parent = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
        loader = factory.create(BaseConfig(os.path.join(parent, 'tests', 'lib', '_root.yml')), 'file')
        items = loader.load({'file': 'var-loader-app/dummy.pem'})
        self.assertIsNotNone(items.get(''))
        self.assertIn('BEGIN', items.get(''))



import os
from unittest import TestCase

import yaml

from octoploy.config.Config import RootConfig, RunMode
from octoploy.deploy.AppDeploy import AppDeployment
from octoploy.utils.Errors import MissingParam
from octoploy.utils.Yml import Yml
from tests import TestUtils


class AppDeploymentTest(TestCase):

    def setUp(self) -> None:
        self._base_path = os.path.dirname(__file__)
        self._tmp_file = 'out.yml'
        self._mode = RunMode()
        self._mode.dry_run = True
        self._mode.out_file = self._tmp_file

    def tearDown(self) -> None:
        if os.path.isfile(self._tmp_file):
            os.remove(self._tmp_file)

    def test_inherit_vars(self):
        """
        Makes sure variables from the app are passed to the library
        and overrides conflicts with values from the app
        """
        self._deploy('app')
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('ABC', data['metadata']['name'])
        self.assertEqual('ABC-1', data['metadata']['REMAPPED'])
        self.assertEqual('ABC', data['metadata']['REMAPPED2'])
        self.assertEqual('DEF', data['metadata']['name2'])
        self.assertEqual('3', data['metadata']['base'])
        # Global variable should be passed down and should not be
        # overwritten by apps
        self.assertEqual('global', data['metadata']['GLOBAL_TEST'])

    def test_var_override(self):
        self._mode.var_override['globalVar'] = 'ext-override'
        self._deploy('app')
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('ext-override', data['metadata']['GLOBAL_TEST'])

    def test_cm_types(self):
        self._deploy('cm-types')

        with open(self._tmp_file) as f:
            content = f.read()
        # Make sure the "y" is quoted
        self.assertIn('"y"', content)

    def test_library(self):
        self._deploy(None, project='lib-usage')

        with open(self._tmp_file) as f:
            data = list(yaml.load_all(f, Loader=yaml.FullLoader))
        data.sort(key=lambda d: d['metadata']['name'])
        self.assertEqual(2, len(data))
        self.assertEqual('paramValue', data[1]['metadata']['name'])

    def test_library_inherit_app_flags(self):
        self._deploy(None, project='lib-usage-flags')

        with open(self._tmp_file) as f:
            data = list(yaml.load_all(f, Loader=yaml.FullLoader))
        self.assertEqual(1, len(data))

    def test_var_loader(self):
        self._deploy('var-loader-app', project='lib-usage')

        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        expected = '''-----BEGIN PRIVATE KEY-----
KEY STUFF
-----END PRIVATE KEY-----
'''
        self.assertEqual(expected, data['KEY'])

    def test_for_each(self):
        self._deploy('app-for-each')

        docs = Yml.load_docs(self._tmp_file)

        # We should have two instances
        self.assertEqual(2, len(docs))
        self.assertEqual('hello', docs[0]['metadata']['REMAPPED'])
        self.assertEqual('hello', docs[1]['metadata']['REMAPPED'])

    def test_params(self):
        prj_config = RootConfig.load(os.path.join(self._base_path, 'app_deploy_test_params'))
        app_config = prj_config.load_app_config('app-params')
        runner = AppDeployment(prj_config, app_config, self._mode)
        try:
            runner.deploy()
            self.fail('No exception raised for missing param')
        except MissingParam:
            pass

        external_params = {
            'SomeParam': 'input'
        }
        app_config._external_vars = external_params
        runner.deploy()

        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('input', data['someObj']['param'])

    def test_secret_without_encryption(self):
        """
        Makes s ure secrets without encrypted values are not deployed
        """
        os.environ['OCTOPLOY_KEY'] = TestUtils.OCTOPLOY_KEY
        self._deploy('secrets')
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertTrue('plainText' not in data['stringData'])

    def test_decryption(self):
        os.environ['OCTOPLOY_KEY'] = TestUtils.OCTOPLOY_KEY
        self._deploy('secrets')
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('hello world', data['stringData']['ref'])
        self.assertEqual('hello world', data['stringData']['field'])

    def _deploy(self, app: str, project: str = 'app_deploy_test'):
        prj_config = RootConfig.load(os.path.join(self._base_path, project))
        prj_config.initialize_state(self._mode)
        if app is None:
            configs = prj_config.load_app_configs()
            for app_config in configs:
                runner = AppDeployment(prj_config, app_config, self._mode)
                runner.deploy()
            return
        app_config = prj_config.load_app_config(app)
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()



import os.path
import re
from typing import Dict

import yaml

from octoploy.utils.DictUtils import DictUtils


class HelmToOcto:
    """
    Converts a rendered helm chart (single yml file) to an octoploy project
    """

    MANAGED_BY = 'app.kubernetes.io/managed-by'

    def __init__(self, dest: str):
        self._filter = []
        self._dest = dest
        if not os.path.exists(dest):
            raise FileNotFoundError(f"Destination directory {dest} does not exist")

    def include(self, app_name: str):
        self._filter.append(app_name)

    def convert(self, source: str):
        with open(source) as f:
            data = yaml.load_all(f, Loader=yaml.FullLoader)
            for doc in data:
                if doc is None:
                    continue
                self._convert_doc(doc)

    def _convert_doc(self, doc: Dict[str, any]):
        app_name = 'misc'
        labels = DictUtils.get(doc, 'metadata.labels')
        if labels is not None:
            app_name = labels.get('app.kubernetes.io/component', app_name)
            if self.MANAGED_BY in labels:
                del labels[self.MANAGED_BY]

        if not self._include_app(app_name):
            return
        self._add_to_app(doc, app_name)

    def _add_to_app(self, doc: Dict[str, any], app_name: str):
        app_name = app_name.replace(' ', '-').lower()
        dest = os.path.join(self._dest, app_name)
        if not os.path.exists(dest):
            os.mkdir(dest)
            meta_data = {
                'name': app_name,
            }
            with open(os.path.join(dest, '_index.yml'), 'w') as f:
                yaml.dump(meta_data, f)

        kind = doc.get('kind')
        if kind is None:
            # Empty object?
            return

        file_name = re.sub(r'(?<!^)(?=[A-Z])', '-', kind).lower()
        file_path = os.path.join(dest, file_name + '.yml')
        data = [doc]
        if os.path.exists(file_path):
            # Append
            with open(file_path) as f:
                data = list(yaml.load_all(f, Loader=yaml.FullLoader))
            data.append(doc)

        with open(file_path, 'w') as f:
            yaml.dump_all(data, f)

    def _include_app(self, app_name: str) -> bool:
        if len(self._filter) == 0:
            return True
        return app_name in self._filter



from unittest import TestCase, mock

from octoploy.config.Config import AppConfig
from octoploy.k8s.BaseObj import BaseObj
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor


class YmlTemplateProcessorTest(TestCase):

    def test_keep(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
vars:
    MY_VAR: testVal
    IMAGE_NAME: image
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': 'Test',
            'apiVersion': 'v1',
            'root': {
                '1': '$',
                '2': '$$',
                '3': 'item$',
                '4': 'my$var',
                '5': '$item',
                '6': '${item',
                '7': '${',
                '8': '${nonExistent}',
                '9': '${nonExistent}Other${MY_VAR}',
            }}
        proc.process(BaseObj(data))
        self.assertEqual('$', data['root']['1'])
        self.assertEqual('$', data['root']['2'])
        self.assertEqual('item$', data['root']['3'])
        self.assertEqual('my$var', data['root']['4'])
        self.assertEqual('$item', data['root']['5'])
        self.assertEqual('${item', data['root']['6'])
        self.assertEqual('${', data['root']['7'])
        self.assertEqual('${nonExistent}', data['root']['8'])
        self.assertEqual('${nonExistent}OthertestVal', data['root']['9'])

    def test_replace(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
vars:
    MY_VAR: testVal
    IMAGE_NAME: image
    NUMBER: 10
    MY_OBJECT:
        someItem: 1
        someOtherItem: 2
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': 'Test',
            'apiVersion': 'v1',
            'root': {
                'item': '${APP_NAME}',
                'number': '${NUMBER}',
                'object': '${MY_OBJECT}',
                'list': [{
                    'other': '${APP_NAME}/${IMAGE_NAME}'
                }],
                'sub': {
                    'item2': '${MY_VAR}'
                }
            }}
        proc.process(BaseObj(data))
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual(10, data['root']['number'])
        self.assertEqual('hello/image', data['root']['list'][0]['other'])
        self.assertEqual('testVal', data['root']['sub']['item2'])
        self.assertEqual({
            'someItem': 1,
            'someOtherItem': 2
        }, data['root']['object'])

    def test_params(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
vars:
    MY_VAR: testVal
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': 'Test',
            'apiVersion': 'v1',
            'root': {
                'item': '${APP_NAME}',
                'list': [{
                    'other': '${APP_NAME}'
                }],
                'sub': {
                    'item2': '${MY_VAR}'
                }
            }}
        proc.process(BaseObj(data))
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('hello', data['root']['list'][0]['other'])
        self.assertEqual('testVal', data['root']['sub']['item2'])

    def test_escaping(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
vars:
    MY_VAR: testVal
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': '$${APP_NAME}${APP_NAME}Test${APP_NAME}',
            'apiVersion': ''
        }
        proc.process(BaseObj(data))
        self.assertEqual('${APP_NAME}helloTesthello', data['kind'])

    def test_merge_var_inline(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
vars:
    MERGE_OBJ: 
        someKey: value
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': 'Test',
            'apiVersion': 'v1',
            'root': {
                'item': '${APP_NAME}',
                '_merge': '${MERGE_OBJ}',
            }}
        proc.process(BaseObj(data))
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('value', data['root']['someKey'])

    def test_merge_object_inline(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
name: hello
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {
            'kind': 'Test',
            'apiVersion': 'v1',
            'root': {
                'item': '${APP_NAME}',
                '_merge': {
                    'someKey': 'value',
                },
            }}
        proc.process(BaseObj(data))
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('value', data['root']['someKey'])



import difflib
from typing import Dict, List, Iterator, Set

from octoploy.api.Kubectl import K8sApi
from octoploy.k8s.BaseObj import BaseObj
from octoploy.utils.Log import ColorFormatter
from octoploy.utils.YmlWriter import YmlWriter


class K8sObjectDiff:
    """
    Creates nice to look at diffs of two yml files.
    This diff ignores any changes injected by k8s itself (e.g. status, timestamps, ..)
    """

    def __init__(self, k8s: K8sApi):
        self._api = k8s

    def print(self, current: BaseObj, new: BaseObj):
        """
        Prints a diff
        :param current: The current object in the cluster
        :param new: The new object
        """
        mask_values: List[List[str]] = []
        if current.is_kind('secret') or new.is_kind('secret'):
            mask_values.append(['spec', 'data'])
            mask_values.append(['spec', 'stringData'])

        current_data = self._filter_injected(current.data)

        # Server side dry-run to get the same format / list sorting
        new = self._api.dry_run(YmlWriter.dump(new.data))
        new_data = self._filter_injected(new.data)
        self._print_diff(current_data, new_data, [], mask_values)

    def _print_diff(self, current_data: Dict[str, any], new_data: Dict[str, any],
                    context: List[str], mask_values: List[List[str]]):
        if current_data is None:
            current_data = {}
        if new_data is None:
            new_data = {}
        all_keys = set(current_data.keys())
        all_keys.update(new_data.keys())
        all_keys = sorted(all_keys)

        for key in all_keys:
            current_entry = current_data.get(key)
            new_entry = new_data.get(key)
            self._print_value_diff(current_entry, new_entry, context + [key], mask_values)

    def _print_value_diff(self, current_entry, new_entry, context: List[str], mask_values: List[List[str]]):
        mask_value: bool = context in mask_values

        if isinstance(current_entry, list) or isinstance(new_entry, list):
            if current_entry is None:
                current_entry = []
            if new_entry is None:
                new_entry = []

            count = max(len(current_entry), len(new_entry))
            for i in range(count):
                current_val = None
                new_val = None
                if i < len(current_entry):
                    current_val = current_entry[i]
                if i < len(new_entry):
                    new_val = new_entry[i]
                self._print_value_diff(current_val, new_val, context + [f'[{i}]'], mask_values)
            return

        if isinstance(current_entry, dict) or isinstance(new_entry, dict):
            self._print_diff(current_entry, new_entry, context, mask_values)
            return
        if current_entry == new_entry:
            return

        if mask_value:
            if current_entry is not None:
                current_entry = '***'
            if new_entry is not None:
                new_entry = '***'

        if current_entry is None:
            # The entire tree will be added
            print(ColorFormatter.colorize(
                '+ ' + '.'.join(context) + f' = {new_entry}',
                ColorFormatter.green))
            return
        if new_entry is None:
            # The entire tree will be removed
            print(ColorFormatter.colorize(
                '- ' + '.'.join(context) + f' = {current_entry}',
                ColorFormatter.red))
            return

        # If the entry is a multiline string, we create a proper diff
        if (isinstance(current_entry, str) and isinstance(new_entry, str) and
                ('\n' in current_entry or '\n' in new_entry)):
            delta = difflib.unified_diff(current_entry.splitlines(), new_entry.splitlines())
            print(ColorFormatter.colorize(
                '~ ' + '.'.join(context) + f':',
                ColorFormatter.yellow))
            self._print_text_diff(delta)
            return

        print(ColorFormatter.colorize(
            '~ ' + '.'.join(context) + f' = {current_entry} -> {new_entry}',
            ColorFormatter.yellow))

    def _filter_injected(self, data: Dict[str, any]) -> Dict[str, any]:
        """Removes injected fields"""
        data = dict(data)  # Create a copy
        metadata = data.get('metadata', {})
        self._del(metadata, 'creationTimestamp')
        self._del(metadata, 'generation')
        self._del(metadata, 'resourceVersion')
        self._del(metadata, 'uid')
        self._del(metadata, 'finalizers')
        self._del(metadata.get('annotations', {}), 'kubectl.kubernetes.io/last-applied-configuration')
        self._del(metadata.get('annotations', {}), 'kubectl.kubernetes.io/restartedAt')
        self._del(data, 'status')
        return data

    @staticmethod
    def _del(data: Dict[str, any], key: str):
        if key in data:
            del data[key]

    @staticmethod
    def _print_text_diff(delta: Iterator[str]):
        for line in delta:
            line = line.strip('\n').strip()
            if line == '---' or line == '+++':
                continue
            color = ''
            if line.startswith('+'):
                color = ColorFormatter.green
            elif line.startswith('-'):
                color = ColorFormatter.red
            print(ColorFormatter.colorize('\t' + line, color))



from __future__ import annotations

import os
from typing import Optional, Dict, List

from octoploy.api.Kubectl import Oc, K8s, K8sApi
from octoploy.config.AppConfig import AppConfig
from octoploy.config.BaseConfig import BaseConfig
from octoploy.processing import Constants
from octoploy.processing.DataPreProcessor import DataPreProcessor, OcToK8PreProcessor
from octoploy.processing.DecryptionProcessor import DecryptionProcessor
from octoploy.processing.NamespaceProcessor import NamespaceProcessor
from octoploy.processing.TreeWalker import TreeProcessor
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor
from octoploy.state.StateTracking import StateTracking
from octoploy.utils.Errors import ConfigError
from octoploy.utils.Log import Log


class RunMode:
    def __init__(self):
        self.out_file = None  # type: Optional[str]
        """
        Yml output file
        """

        self.dry_run = False
        """
        True if no OC should be called
        """

        self.plan = False
        """
        True if changes should be previewed
        """

        self.var_override: Dict[str, str] = {}
        """
        Key/value pairs which should be passed to the templating engine.
        These value have priority over already defined values
        """

        self.delete = False
        """
        Delete mode
        """

    def set_override_env(self, env: List[str]):
        """
        Parses a key=value list
        :param env: key=value strings
        """
        for entry in env:
            parts = entry.split('=', 1)
            if len(parts) < 2:
                raise ValueError(f'Could not parse {entry}')
            key = parts[0]
            value = parts[1]
            self.var_override[key] = value


class RootConfig(BaseConfig):
    """
    Root configuration for a project (aka a collection of apps inside a single context/namespace)
    """

    _app_flags: Dict[str, Dict[str, any]]
    """
    Holds any override parameters for 
    """

    _parent: Optional[RootConfig] = None
    """
    The parent that uses this config.
    This may only be set if this config is a library.
    """

    _libraries: List[RootConfig]
    """
    List of libraries that should be inherited into this config
    """

    def __init__(self, config_root: str, path: str):
        super().__init__(path)
        self.log = Log(__name__).log
        self._config_root = config_root
        self._k8s_api = None
        self._libraries = []
        self._global_var_overrides: Dict[str, str] = {}

        parent_dir = os.path.abspath(os.path.join(path, os.pardir, os.pardir))
        inherit = self.data.get('inherit')
        if inherit is not None:
            self.log.warning('"inherit" field is deprecated, use "libraries" instead')
            lib_dir = os.path.join(parent_dir, inherit)
            self._load_library(lib_dir)

        for lib in self.data.get('libraries', []):
            lib_dir = os.path.join(parent_dir, lib)
            self._load_library(lib_dir)

        state_name = self.data.get('stateName', '')
        self._state = StateTracking(self.create_api(), state_name)

        self._app_flags = self.data.get('apps', {})

    def get_var_overrides(self) -> Dict[str, str]:
        return self._global_var_overrides

    @classmethod
    def load(cls, path: str) -> RootConfig:
        return RootConfig(path, os.path.join(path, '_root.yml'))

    def app_is_enabled(self, name: str) -> bool:
        """
        Indicates if the given app should be disabled
        :param name: Name of the app
        :return: Disabled
        """
        if self._parent is not None and not self._parent.app_is_enabled(name):
            return False

        flags = self._app_flags.get(name, {})
        return flags.get('enabled', True)

    def get_state(self) -> StateTracking:
        return self._state

    def initialize_state(self, run_mode: RunMode):
        self._global_var_overrides = run_mode.var_override
        if run_mode.out_file is not None:
            if os.path.isfile(run_mode.out_file):
                os.remove(run_mode.out_file)
        if run_mode.dry_run:
            return
        self._state.restore(self.get_namespace_name())

    def persist_state(self, run_mode: RunMode):
        if run_mode.dry_run or run_mode.plan:
            return
        self._state.store(self.get_namespace_name())

    def get_config_root(self) -> str:
        return self._config_root

    def is_library(self) -> bool:
        """
        Indicates if this collection is a library
        """
        return self.data.get('type', '') == 'library'

    def _get_mode(self) -> str:
        return self.data.get('mode', 'k8s')

    def create_api(self) -> K8sApi:
        """
        Creates a new openshift / k8s client.
        :return: Client
        """
        if self._k8s_api is not None:
            return self._k8s_api

        mode = self._get_mode()
        if mode == 'oc':
            k8s_api = Oc()
        elif mode == 'k8s' or mode == 'k8':
            k8s_api = K8s()
        else:
            raise ValueError(f'Invalid mode: {mode}')

        context = self.get_kubectl_context()
        if context is not None:
            k8s_api.switch_context(context)
        self._k8s_api = k8s_api
        return k8s_api

    def get_namespace_name(self) -> Optional[str]:
        """
        Returns the namespace name of this project

        :return: Name or null for libraries or if the namespace should not be changed
        """
        namespace = self.data.get('namespace', self.data.get('project'))
        if namespace is None:
            # Maybe passed via commandline?
            return self._global_var_overrides.get(Constants.VAR_NAMESPACE)
        return namespace

    def get_kubectl_context(self) -> Optional[str]:
        """
        Returns the configuration context name.
        :return: Name or null if the current context should be used
        """
        return self.data.get('context')

    def get_pre_processor(self) -> DataPreProcessor:
        """
        Returns the pre-processor for the current config
        """
        mode = self._get_mode()
        if mode == 'k8s' or mode == 'k8s':
            return OcToK8PreProcessor()
        return DataPreProcessor()

    def get_template_processor(self) -> YmlTemplateProcessor:
        root_processor = super().get_template_processor()
        parent_processors = []
        for lib in self._libraries:
            parent_processors.append(lib.get_template_processor())
        root_processor.parents(parent_processors)
        return root_processor

    def get_replacements(self) -> Dict[str, str]:
        """
        Returns all variables which are available for the yml files
        :return: Key, value map
        """
        items = super().get_replacements()
        name = self.get_namespace_name()
        if name is not None:
            items.update({
                'OC_PROJECT': name,
                Constants.VAR_NAMESPACE: name
            })

        items.update(self._global_var_overrides)
        return items

    def get_yml_processors(self) -> List[TreeProcessor]:
        """
        Returns all yml processors which should be applied
        :return: Processors
        """
        return [DecryptionProcessor(), NamespaceProcessor(self)]

    def load_app_configs(self) -> List[AppConfig]:
        """
        Loads all app configurations available in this project
        :return:
        """
        items = []
        names = set()
        for dir_item in os.listdir(self._config_root):
            path = os.path.join(self._config_root, dir_item)
            if not os.path.isdir(path):
                continue
            try:
                app_config = self.load_app_config(dir_item)
            except FileNotFoundError:
                # Index file missing
                continue

            app_name = app_config.get_name()
            if app_config.is_template() or not app_config.enabled():
                self.log.debug(f"Skipping {app_name} as it's disabled")
                continue

            if app_name in names:
                raise ValueError(f'The app name {app_name} has already been used')
            names.add(app_name)
            items.append(app_config)
        for library in self._libraries:
            items.extend(library.load_app_configs())
        return items

    def load_app_config(self, name: str) -> AppConfig:
        folder_path = os.path.join(self._config_root, name)
        if not os.path.isdir(folder_path):
            # Search in library
            for library in self._libraries:
                try:
                    return library.load_app_config(name)
                except FileNotFoundError:
                    continue
            raise FileNotFoundError('App folder not found: ' + folder_path)

        index_file = os.path.join(folder_path, '_index.yml')
        if not os.path.isfile(index_file):
            # Index file missing
            raise FileNotFoundError('No index yml file found: ' + index_file)

        variables = self.get_replacements()
        return AppConfig(folder_path, index_file, variables, root=self)

    def _load_library(self, lib_dir: str):
        """
        Loads the library project at the given directory
        :param lib_dir: Path to the root of the library directory
        """
        if not os.path.isdir(lib_dir):
            raise FileNotFoundError('Library not found: ' + lib_dir)
        library = RootConfig.load(lib_dir)
        library._parent = self
        if not library.is_library():
            raise ConfigError(f'{lib_dir} referenced but is not marked as library')
        self._libraries.append(library)
